from starlette.responses import RedirectResponse
from oauthlib.common import generate_token
from .oauth import OAuth2Session
from .models import User, Guild, Connection


DISCORD_URL = 'https://discord.com'
API_URL = DISCORD_URL + '/api/v9'


class DiscordOAuthSession(OAuth2Session):
    """Session containing data for a single authorized user. Handles authorization internally.

    .. warning::
        It is recommended to not construct this class directly.
        Use `DiscordOAuthClient.session` or `DiscordOAuthClient.session_from_token` instead.

    .. note::
        Either the 'code' or 'token' parameter must be provided, but not both.

    Parameters
    ----------
    code: Optional[:class:`str`]
        Authorization code included with user request after redirect from Discord.
    token: Optional[Dict[:class:`str`, Union[:class:`str`, :class`int`, :class:`float`]]]
        A previously generated, valid, access token to use instead of the OAuth code exchange
    client_id: :class:`str`
        Your Discord application client ID.
    scope: :class:`str`
        Discord authorization scopes separated by %20.
    redirect_uri: :class:`str`
        Your Discord application redirect URI.
    code: Optional[:class:`str`]
        Authorization code included with user request after redirect from Discord.
    token: Optional[Dict[:class:`str`, Union[:class:`str`, :class:`int`, :class`float`]]]
        A previously generated, valid, access token to use instead of the OAuth code exchange
    """
    def __init__(self, client_id, client_secret, scope, redirect_uri, *, code, token):
        if (not (code or token)) or (code and token):
            raise ValueError("Either 'code' or 'token' parameter must be provided, but not both.")
        elif token:
            if not isinstance(token, dict):
                raise TypeError("Parameter 'token' must be an instance of dict with at least the 'access_token' key.'")
            if 'access_token' not in token:
                raise ValueError("Parameter 'token' requires 'access_token' key.")
            elif not token.get('token_type'):  # this is not required for the discord class but for the parent class
                token['token_type'] = 'Bearer'

        self._discord_auth_code = code
        self._discord_client_secret = client_secret
        self._discord_token = token
        self._cached_user = None
        self._cached_guilds = None
        self._cached_connections = None

        super().__init__(
            client_id=client_id,
            scope=scope,
            redirect_uri=redirect_uri,
            token=token
        )

    @property
    def token(self):
        """Dict[:class:`str`, Union[:class:`str`, :class:`int`, :class:`float`]]: The session's current OAuth token, if one exists."""
        # this is pretty much just for documentation purposes.
        return super().token

    @token.setter
    def token(self, value):
        # stolen from oauth.py to allow the client to set this still.
        self._client.token = value
        self._client.populate_token_attributes(value)

    @property
    def cached_user(self):
        """:class:`dict`: The session's cached user, if an `identify()` request has previously been made."""
        return self._cached_user

    @property
    def cached_guilds(self):
        """List[:class:`dict`]: The session's cached guilds, if a `guilds()` request has previously been made."""
        return self._cached_guilds

    @property
    def cached_connections(self):
        """List[:class:`dict`]: The session's cached account connections, if a `connections()` request has previously been made."""
        return self._cached_connections

    def new_state(self):
        """Generate a new state string for verifying authorizations.

        Returns
        -------
        :class:`str`
            The state string that was generated.
        """
        return generate_token()

    async def _discord_request(self, url_fragment, method='GET'):
        if not self._discord_token:
            url = API_URL + '/oauth2/token'
            self._discord_token = await self.fetch_token(
                url,
                code=self._discord_auth_code,
                client_secret=self._discord_client_secret
            )

        token = self._discord_token['access_token']
        url = API_URL + url_fragment
        headers = {
            'Authorization': 'Authorization: Bearer ' + token
        }
        async with self.request(method, url, headers=headers) as resp:
            resp.raise_for_status()
            return await resp.json()

    async def identify(self):
        """Identify a user.

        Returns
        -------
        :class:`User`
            The user who authorized the application.
        """
        data_user = await self._discord_request('/users/@me')
        user = User(data=data_user)
        self._cached_user = user
        return user

    async def guilds(self):
        """Fetch a user's guild list.

        Returns
        -------
        List[:class:`Guild`]
            The user's guild list.
        """
        data_guilds = await self._discord_request('/users/@me/guilds')
        guilds = [Guild(data=g) for g in data_guilds]
        self._cached_guilds = guilds
        return guilds

    async def connections(self):
        """Fetch a user's linked 3rd-party accounts.

        Returns
        -------
        List[:class:`Connection`]
            The user's connections.
        """
        data_connections = await self._discord_request('/users/@me/connections')
        connections = [Connection(data=c) for c in data_connections]
        self._cached_connections = connections
        return connections

    async def join_guild(self, guild_id, user_id=None):
        """Add a user to a guild.

        Parameters
        ----------
        guild_id: :class:`int`
            The ID of the guild to add the user to.
        user_id: Optional[:class:`int`]
            ID of the user, if known. If not specified, will first identify the user.
        """
        if not user_id:
            user = await self.identify()
            user_id = user.id
        return await self._discord_request(f'/guilds/{guild_id}/members/{user_id}', method='PUT')

    async def join_group_dm(self, dm_channel_id, user_id=None):
        """Add a user to a group DM.

        Parameters
        ----------
        dm_channel_id: :class:`int`
            The ID of the DM channel to add the user to.
        user_id: Optional[:class:`int`]
            ID of the user, if known. If not specified, will first identify the user.
        """
        if not user_id:
            user = await self.identify()
            user_id = user.id
        return await self._discord_request(f'/channels/{dm_channel_id}/recipients/{user_id}', method='PUT')


class DiscordOAuthClient:
    """Client for Discord Oauth2.

    Parameters
    ----------
    client_id: Union[:class:`str`, :class:`int`]
        Discord application client ID.
    client_secret: :class:`str`
        Discord application client secret.
    redirect_uri: :class:`str`
        Discord application redirect URI.
    scopes: Tuple[:class:`str`]
        Discord authorization scopes.
    """

    def __init__(self, client_id, client_secret, redirect_uri, scopes=('identify',)):
        self.client_id = str(client_id)
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.scope = ' '.join(scope for scope in scopes)

    def redirect(self, state=None, prompt=None, redirect_uri=None):
        """Returns a RedirectResponse that directs to Discord login.

        Parameters
        ----------
        state: Optional[:class:`str`]
            Optional state parameter for Discord redirect URL.
            Docs can be found `here <https://discord.com/developers/docs/topics/oauth2#state-and-security>`_.
        prompt: Optional[:class:`str`]
            Optional prompt parameter for Discord redirect URL.
            If ``consent``, user is prompted to re-approve authorization. If ``none``, skips authorization if user has already authorized.
            Defaults to ``consent``.
        redirect_uri: Optional[:class:`str`]
            Optional redirect URI to pass to Discord. Defaults to the client's redirect URI.
        """
        client_id = f'client_id={self.client_id}'
        redirect_uri = f'redirect_uri={redirect_uri or self.redirect_uri}'
        scopes = f'scope={self.scope}'
        response_type = 'response_type=code'
        url = DISCORD_URL + f'/api/oauth2/authorize?{client_id}&{redirect_uri}&{scopes}&{response_type}'
        if state:
            url += f'&state={state}'
        if prompt:
            url += f'&prompt={prompt}'
        return RedirectResponse(url)

    def session(self, code) -> DiscordOAuthSession:
        """Create a new DiscordOAuthSession from an authorization code.

        Parameters
        ----------
        code: :class:`str`
            The OAuth2 code provided by the Discord API.

        Returns
        -------
        :class:`DiscordOAuthSession`
            A new OAuth session.
        """
        return DiscordOAuthSession(
            code=code,
            token=None,
            client_id=self.client_id,
            client_secret=self.client_secret,
            scope=self.scope,
            redirect_uri=self.redirect_uri,
        )

    def session_from_token(self, token) -> DiscordOAuthSession:
        """Create a new DiscordOAuthSession from an existing token.

        Parameters
        ----------
        token: Dict[:class:`str`, Union[:class:`str`, :class:`int`, :class:`float`]]
            An existing (valid) access token to use instead of the OAuth code exchange.

        Returns
        -------
        :class:`DiscordOAuthSession`
            A new OAuth session.
        """
        return DiscordOAuthSession(
            code=None,
            token=token,
            client_id=self.client_id,
            client_secret=self.client_secret,
            scope=self.scope,
            redirect_uri=self.redirect_uri,
        )

    async def login(self, code):
        """Shorthand for session setup + identify.
        
        Parameters
        ----------
        code: :class:`str`
            The OAuth2 code provided by the authorization request.
        
        Returns
        -------
        :class:`User`
            The user who authorized the application.
        """
        async with self.session(code) as session:
            user = await session.identify()
        return user

    # TODO: decide if I want to keep this. not a fan of the method name.
    async def login_return_token(self, code):
        """Shorthand for session setup + identify. Returns user and token.
        
        Parameters
        ----------
        code: :class:`str`
            The OAuth2 code provided by the authorization request.

        Returns
        -------
        Tuple[user, token]
        user: :class:`User`
            The user who authorized the application.
        token: Dict[:class:`str`, Union[:class:`str`, :class:`int`, :class:`float`]]
            The access token provided by the Discord API.
        """
        async with self.session(code) as session:
            user = await session.identify()
        return user, session.token



from typing import Optional, List
import discord


class DiscordObject:
    """Represents a Discord object. This library's equivalent to discord.Object.

    Attributes
    ----------
    id: :class:`int`
        The Discord object's unique ID.
    """

    def __init__(self, data):
        self._json_data = data
        self.id = int(data['id'])

    @classmethod
    def from_id(cls, id_: int):
        """Initializes a new DiscordObject with the given ID.

        .. note::
            Most people will never have to use this method.
            It is provided for cases where you only need an object with a specific ID.

        Parameters
        ----------
        id_: :class:`int`
            The ID of the object to create.
        """
        return cls({'id': str(id_)})

    def __eq__(self, other) -> bool:
        return other.id == self.id

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return self.id >> 22

    def json(self):
        """Returns the original JSON data for this model."""
        return self._json_data


class User(DiscordObject):
    """A `user`_ model from Discord. Returned by ``session.identify()``.

    Attributes
    ----------
    id: :class:`int`
        The user's unique ID.
    username: :class:`str`
        The user's username.
    discriminator: :class:`str`
        The user's discriminator.
    avatar: :class:`str`
        The user's avatar hash.
    flags: :class:`int`
        The `flags`_ on the user's account
    public_flags: :class:`int`
        The public `flags`_ on the user's account.
    banner: :class:`str`
        The user's banner hash.
    banner_color: :class:`str`
        The user's banner color.
    accent_color: :class:`int`
        The user's banner color: represented as the int form of the color's hex code.
    locale: :class:`str`
        The user's language locale.
    mfa_enabled: :class:`bool`
        Whether the user has multi-factor authentication enabled.
    email: :class:`str`
        The user's email address. Only provided if the ``email`` scope is authorized.
    verified: :class:`str`
        Whether the user's email address has been verified.


    .. _user: https://discord.com/developers/docs/resources/user
    .. _flags: https://discord.com/developers/docs/resources/user#user-object-user-flags
    """

    __slots__ = (
        '_json_data',
        'id',
        'username',
        'discriminator',
        'avatar',
        'flags',
        'public_flags'
        'banner',
        'banner_color'
        'accent_color',
        'locale',
        'mfa_enabled',
        'email',
        'verified',
    )

    _json_data: dict
    id: int
    username: str
    discriminator: str
    avatar: Optional[str]
    flags: int
    public_flags: int
    banner: Optional[str]
    banner_color: Optional[int]
    accent_color: Optional[int]
    locale: Optional[str]
    mfa_enabled: bool
    email: Optional[str]
    verified: bool

    def __init__(self, *, data):
        super().__init__(data)
        self._update(data)

    def __repr__(self) -> str:
        return f"<User id={self.id} username={self.username!r} discriminator={self.discriminator!r}>"

    def __str__(self) -> str:
        return f'{self.username}#{self.discriminator}'

    def _update(self, data) -> None:
        # these are defined in super().__init__
        # self._json_data = data
        # self.id = int(data['id'])
        self.username = data['username']
        self.discriminator = data['discriminator']
        self.avatar = data['avatar']
        self.flags = data['flags']
        self.public_flags = data.get('public_flags', 0)
        self.banner = data.get('banner', None)
        self.banner_color = data.get('banner_color', None)
        self.accent_color = data.get('accent_color', None)
        self.locale = data.get('locale', None)
        self.mfa_enabled = data.get('mfa_enabled', None)
        self.email = data.get('email', None)
        self.verified = data.get('verified', None)

    async def to_dpy(self, client):
        """Tries to convert this User to a ``discord.User``.

        This is just a shortcut for ``client.get_user(id)`` followed by ``client.fetch_user(id)``,
        returning ``None`` if not found.

        .. note::
            A discord.py ``Client`` or ``Bot`` object must be passed into this function.

        Parameters
        ----------
        client: :class:`discord.Client`
            The bot client to use to create the object.

        Returns
        -------
        :class:`discord.User`
            The discord.py User object, if it could be found.
        """
        user = client.get_user(self.id)
        if not user:
            try:
                user = await client.fetch_user(self.id)
            except discord.HTTPException:
                return None
        return user


class Guild(DiscordObject):
    """A partial `guild`_ model from Discord. Returned by ``session.guilds()``.

    Attributes
    ----------
    id: :class:`int`
        The guild's unique ID.
    name: :class:`str`
        The guild's name.
    icon: :class:`str`
        The guild's icon hash.
    owner: :class:`bool`
        Whether the authorized user is owner of this guild.
    permissions: :class:`int`
        The authorized user's `permissions`_ in this guild.
    features: List[:class:`str`]
        The guild's enabled `features`_.


    .. _guild: https://discord.com/developers/docs/resources/guild
    .. _features: https://discord.com/developers/docs/resources/guild#guild-object-guild-features
    .. _permissions: https://discord.com/developers/docs/topics/permissions
    """

    __slots__ = (
        '_json_data',
        'id',
        'name',
        'icon',
        'owner',
        'permissions',
        'features',
    )

    _json_data: dict
    id: int
    name: str
    icon: Optional[str]
    owner: bool
    permissions: int
    features: List[str]

    def __init__(self, *, data):
        self._update(data)
        super().__init__(data)

    def __repr__(self) -> str:
        return f"<Guild id={self.id} name={self.name!r}>"

    def __str__(self) -> str:
        return self.__repr__()

    def _update(self, data) -> None:
        # these are defined in super().__init__
        # self._json_data = data
        # self.id = int(data['id'])
        self.name = data['name']
        self.icon = data.get('icon', None)
        self.owner = data['owner']
        self.permissions = int(data['permissions'])
        self.features = data['features']

    async def to_dpy(self, client):
        """Tries to convert this Guild to a ``discord.Guild``.

        This is just a shortcut for ``client.get_guild(id)`` followed by ``client.fetch_guild(id)``,
        returning ``None`` if not found.

        .. note::
            A discord.py ``Client`` or ``Bot`` object must be passed into this function.

        Parameters
        ----------
        client: :class:`discord.Client`
            The bot client to use to create the object.

        Returns
        -------
        :class:`discord.Guild`
            The discord.py Guild object, if the guild could be found.
        """
        guild = client.get_guild(self.id)
        if not guild:
            try:
                guild = await client.fetch_guild(self.id)
            except discord.HTTPException:
                return None
        return guild


class Connection:
    """An account `connection`_ model from Discord.

    Attributes
    ----------
    type: :class:`str`
        The connection type.
    id: :class:`str`
        The connected account's ID.
    name: :class:`str`
        The connected account's username.
    visibility: :class:`int`
        The connected account's `visibility`_.
    friend_sync: :class:`bool`
        Whether friend sync is enabled for this connected account.
    show_activity: :class:`bool`
        Whether activities related to this connection will be shown in presence updates.
    verified: :class:`bool`
        Whether this connected account is verified.


    .. _connection: https://discord.com/developers/docs/resources/user#connection-object
    .. _visibility: https://discord.com/developers/docs/resources/user#connection-object-visibility-types
    """
    
    __slots__ = (
        '_json_data',
        'type',
        'id',
        'name',
        'visibility',
        'friend_sync',
        'show_activity',
        'verified',
    )

    _json_data: dict
    type: str
    id: str
    name: str
    visibility: int
    friend_sync: bool
    show_activity: bool
    verified: bool

    def __init__(self, *, data):
        self._update(data)

    def __repr__(self) -> str:
        return f"<Connection type={self.type}>"

    def __str__(self) -> str:
        return self.__repr__()

    def _update(self, data) -> None:
        self._json_data = data
        self.type = data['type']
        self.id = data['id']
        self.name = data['name']
        self.visibility = data['visibility']
        self.friend_sync = data['friend_sync']
        self.show_activity = data['show_activity']
        self.verified = data['verified']

    def json(self):
        """Returns the original JSON data for this model."""
        return self._json_data



import setuptools
import re


with open('./README.md', 'r') as fp:
    long_description = fp.read()


with open('./starlette_discord/__init__.py', 'r') as fp:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fp.read(), re.MULTILINE).group(1)


with open('requirements.txt') as f:
    requirements = f.read().splitlines()


setuptools.setup(
    name='starlette-discord',
    author='nwunderly',
    url='https://github.com/nwunderly/starlette-discord',
    project_urls={
        "Documentation": "https://starlette-discord.rtfd.io/",
    },
    version=version,
    description='"Login with Discord" support for Starlette and FastAPI.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=requirements,
    extras_require={
        'docs': [
            'sphinx',
            'sphinxcontrib_trio',
            'myst_parser',
        ],
    },
    python_requires='>=3.8',
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
)




"""
starlette-discord
-----------------
"Login with Discord" support for Starlette and FastAPI.

:copyright: (c) 2021 nwunderly
:license: MIT, see LICENSE for more details.
"""

__title__ = 'starlette-discord'
__author__ = 'nwunderly'
__license__ = 'MIT'
__copyright__ = 'Copyright 2021 nwunderly'
__version__ = '0.2.0'

from .client import DiscordOAuthClient, DiscordOAuthSession
from .models import DiscordObject, User, Guild, Connection
